from datetime import datetime
from airflow.models.baseoperator import BaseOperator
from urllib import parse
from acceldata_sdk.events.generic_event import GenericEvent
from acceldata_sdk.models.pipeline import CreatePipeline, PipelineMetadata, PipelineSourceType
from acceldata_sdk.torch_client import TorchClient
from acceldata_sdk.api_version_utils import APIVersionUtils
from acceldata_airflow_sdk.initialiser import torch_credentials
from acceldata_airflow_sdk.utils.torch_client import TorchDAGClient
from acceldata_airflow_sdk.utils.airflow_job_utils import PIPELINE_UID_XCOM, is_pipeline_run_ended, CONNECTION_ID
from semantic_version import Version, SimpleSpec
import logging


LOGGER = logging.getLogger("airflow.task")

def check_if_pipeline_exists(pipeline_uid: str, client):
    try:
        pipeline = client.get_pipeline(pipeline_uid=pipeline_uid)
        LOGGER.info('check_if_pipeline_exists:: pipeline:  %s ', pipeline)
        return True
    except Exception as e:
        LOGGER.warn('check_if_pipeline_exists:: pipeline: False -> ERROR : %s ', str(e))
        return False


class TorchInitializer(BaseOperator):
    template_fields = ['continuation_id']
    """
    You need to add task with given operator at the root of your dag. It will create new pipeline run for your dag run.
    You need to add it as a root of the dag.

    You need to add 6 additional parameters pipeline_uid, pipeline_name, create_pipeline, span_name, continuation_id, meta.
    Other parameters will be same as std airflow base operator's parameters

    """

    def __init__(self, *, pipeline_uid, pipeline_name=None, create_pipeline=True, span_name=None, continuation_id=None,
                 meta: PipelineMetadata = None, connection_id=None, **kwargs):
        """
        You need to add 6 additional parameters pipeline_uid, pipeline_name, create_pipeline, span_name, continuation_id, meta.
        Other parameters will be same as std airflow base operator's parameters

        :param pipeline_uid: (String) uid of the pipeline given in torch
        :param pipeline_name: (String) name of the pipeline given in torch
        :param create_pipeline: (bool) optional False If pipeline, pipeline_run and root span has already been created
            before running Airflow DAG
        :param continuation_id continuation_id of the pipeline run
        :param span_name: (String) optional Custom root span name. If nothing is passed pipeline_uid.span is used as name
        :param meta: (PipelineMetadata) meta data of the pipeline
        :param connection_id connection_id of the connection from where credentials need to be used. If this is None credential will be used from environment variable
        """
        super().__init__(**kwargs)
        self.pipeline_name = pipeline_name
        self.pipeline_uid = pipeline_uid
        self.continuation_id = continuation_id
        self.create_pipeline = create_pipeline
        self.span_name = span_name
        self.connection_id = connection_id
        self.meta = meta
        if self.meta is None:
            self.meta = PipelineMetadata(owner='sdk/pipeline-user', team='TORCH', codeLocation='...')

    def execute(self, context):
        task_instance = context['ti']
        # Make the key unique by appending UUID
        task_instance.xcom_push(key=PIPELINE_UID_XCOM, value=self.pipeline_uid)
        task_instance.xcom_push(key=CONNECTION_ID, value=self.connection_id)

        if self.create_pipeline:
            pipeline_name_ = self.pipeline_uid
            if self.pipeline_name is not None:
                pipeline_name_ = self.pipeline_name
            LOGGER.info('Creating new pipeline with passed uid :: %s',self.pipeline_uid)
            torch_client = TorchClient(**torch_credentials(self.connection_id))
            # Insert the logic to set the sourceType for the pipeline based on the version
            torch_version = torch_client.get_torch_version()
            actual_version = APIVersionUtils.extract_actual_version(torch_version)
            LOGGER.info("Creating pipeline with sourceType as AIRFLOW for the version: %s", actual_version)
            pipeline = CreatePipeline(
                    uid=self.pipeline_uid,
                    name=pipeline_name_,
                    description=f'The pipeline {pipeline_name_} has been created from acceldata-airflow-sdk',
                    meta=self.meta,
                    context={'pipeline_uid': self.pipeline_uid, 'pipeline_name': pipeline_name_},
                    sourceType=PipelineSourceType.AIRFLOW
            )
            pipeline_res = torch_client.create_pipeline(pipeline=pipeline)
            LOGGER.info('pipeline id :: %s', pipeline_res.id)
            if self.continuation_id is not None:
                LOGGER.info(f'Creating new pipeline run with passed continuation_id :: %s {self.continuation_id}.')
                pipeline_run = pipeline_res.create_pipeline_run(continuation_id=self.continuation_id)
            else:
                pipeline_run = pipeline_res.create_pipeline_run()
            task_instance.xcom_push(key=f'{task_instance.dag_id}_pipeline_run_id', value=pipeline_run.id)
            if self.span_name:
                span_name_ = self.span_name
            else:
                span_name_ = f'{self.pipeline_uid}.span'
            parent_span_context = pipeline_run.create_span(uid=span_name_)
        else:
            client = TorchDAGClient(self.connection_id)
            if self.continuation_id is not None:
                LOGGER.info('Using precreated pipeline with continuation_id :: %s ', self.continuation_id)
                pipeline = client.get_pipeline(self.pipeline_uid)
                pipeline_run = client.get_pipeline_run(continuation_id=self.continuation_id, pipeline_id=pipeline.id)
                if is_pipeline_run_ended(pipeline_run):
                    raise Exception("Please provide continuation_id of active pipeline run")
            else:
                LOGGER.info('Using latest pipeline run.')
                pipeline_run = client.get_latest_pipeline_run(self.pipeline_uid)
                if is_pipeline_run_ended(pipeline_run):
                    raise Exception("Latest pipeline run is in terminal state. Please make sure latest pipeline run is not completed")
            parent_span_context = client.get_root_span(pipeline_uid=self.pipeline_uid,
                                                       pipeline_run_id=pipeline_run.id)
            task_instance.xcom_push(key=f'{task_instance.dag_id}_pipeline_run_id', value=pipeline_run.id)
            LOGGER.info('Using precreated pipeline with pipeline uid :: %s ', self.pipeline_uid)
        try:
            log_url = list({context.get('task_instance').log_url})
            list_ = list(log_url)
            url = list_[0]
            parsed = parse.urlsplit(url)
            query = parse.parse_qs(parse.urlsplit(url).query)
            dag_id = query['dag_id'][0]
            execution_date = query['execution_date'][0]
            encoded_time = parse.quote(execution_date)
            dagrun_url = parsed.scheme + '://' + parsed.netloc + '/graph?root=&dag_id=' + dag_id + '&execution_date=' + encoded_time + '&arrang=LR'
            parent_span_context.send_event(GenericEvent(
                context_data={
                    'dag_id': dag_id,
                    'time': str(datetime.now()),
                    'url': dagrun_url,
                    'execution_time': execution_date
                },
                event_uid='AIRFLOW.DETAILS')
            )
        except:
            pass




from airflow.models.baseoperator import BaseOperator
from acceldata_airflow_sdk.utils.torch_client import TorchDAGClient
from acceldata_airflow_sdk.utils.airflow_job_utils import PIPELINE_UID_XCOM, get_dag_run_pipeline_run_id, CONNECTION_ID, \
    create_job_span, log_job_details, get_task_instance, \
    generate_job_uid, handle_span_success, create_context_job, is_span_not_found, create_new_span, \
    handle_span_failure, fetch_xcom_data, send_span_start_generic_event, log_tries, generate_span_operator_uid, \
    pull_xcom_data
from acceldata_sdk.errors import APIError
import logging

LOGGER = logging.getLogger("airflow.task")


class SpanOperator(BaseOperator):
    """
    Description:
        Used to send span start and end event for any std airflow operator. Just wrap your operator with span operator.
        Make sure you do not add your task in dag(dag parameter should not be specified in the operator being wrapped by
         span operator). If you wrap it using span operator, will take care of that task operator.

    You need to add extra parameter mentioned below. Other parameters will be same as std airflow base operator's parameters

    :param operator: std task operator defined
    :param span_uid: span uid for the task
    :param associated_job_uids: list of job uids
    :param xcom_to_event_mapper_ids: list of xcom keys. Used to send xcom variables in span event job

    """

    def __init__(self, *, operator: BaseOperator, span_uid: str = None, associated_job_uids=None,
                 xcom_to_event_mapper_ids=None, **kwargs):
        """
        You need to add extra parameter mentioned below. Other parameters will be same as std airflow base operator's parameters
        :param operator: std task operator defined
        :param span_uid: span uid for the task
        :param associated_job_uids: list of job uids
        :param xcom_to_event_mapper_ids: list of xcom keys. Used to send xcom variables in span event
        Example :

        --> Defined std operator.

        postgres_operator = PostgresOperator(
            task_id="task_name",
            postgres_conn_id='example_db',
            sql="select * from information_schema.attributes",
        )


        --> To wrap operator with span. Write assign this to your dag (not your std operator)

        span_operator = SpanOperator(
            task_id='task_name',
            span_uid='span.uid',
            operator=postgres_operator,
            dag=dag
        )

        """
        if kwargs.get("provide_context"):
            kwargs.pop('provide_context', None)
        super().__init__(**kwargs)
        self.operator = operator
        self.pipeline_uid = None
        self.span_uid = span_uid
        self.parent_span_ctxt = None
        if associated_job_uids is None:
            self.associated_job_uids = []
        else:
            self.associated_job_uids = associated_job_uids
        self.xcom_to_event_mapper_ids = xcom_to_event_mapper_ids
        self.span_context = None

    def execute(self, context):
        xcom_context_data, pipeline_run, try_number, max_retries, child_span_context = (None, None, None, None, None)
        try:
            LOGGER.info("Send span start event")
            task_instance = context['ti']
            parent_span_context = pull_xcom_data(task_instance, 'parent_span_context')
            conn_id = pull_xcom_data(task_instance, CONNECTION_ID)
            try_number = task_instance.try_number
            max_retries = task_instance.task.retries
            log_tries(try_number, max_retries)
            if parent_span_context is None:
                LOGGER.debug('Sending new request to ADOC pipeline service to get parent span context')
                self.pipeline_uid = pull_xcom_data(task_instance, PIPELINE_UID_XCOM)
                client = TorchDAGClient(conn_id)
                pipeline_run_id = get_dag_run_pipeline_run_id(task_instance)
                pipeline = client.get_pipeline(self.pipeline_uid)
                pipeline_run = pipeline.get_run(pipeline_run_id)
                self.parent_span_ctxt = client.get_root_span(pipeline_uid=self.pipeline_uid,
                                                             pipeline_run_id=pipeline_run_id)
                # task_instance.xcom_push(key="parent_span_context", value=self.parent_span_ctxt.__dict__)
            else:
                LOGGER.debug('using xcom to get parent span context to send span event')
                self.parent_span_ctxt = parent_span_context
            if self.span_uid is None:
                LOGGER.debug("Span Uid is None. Generating span uid")
                self.span_uid = generate_span_operator_uid(type(self.operator).__name__, task_instance.task_id)
                LOGGER.debug("Generated Span uid : %s", self.span_uid)

            # Add the logic to check if the span already exists or not and create a new span if span doesn't exist
            try:
                LOGGER.info("Checking if span already exists with the child_span_uid :: %s", self.span_uid)
                child_span_context = pipeline_run.get_span(self.span_uid)
                LOGGER.info("Span already exists with the child_span_uid : %s . Not creating the span again.",
                            self.span_uid)
                # No need to create a span, emit generic events related to the retry here
                send_span_start_generic_event(child_span_context, try_number, self.span_uid)
            except APIError as e:
                if is_span_not_found(e):
                    LOGGER.info("Span does not exist with child_span_uid : %s. Creating new child span. ",
                                self.span_uid)
                    child_span_context = create_new_span(self.parent_span_ctxt, self.span_uid, None, try_number,
                                                         self.associated_job_uids)

            self.span_context = child_span_context
            LOGGER.debug("Span context : %s", self.span_context)
            context['span_context_parent'] = self.span_context
            xcom_context_data = {}
            if self.xcom_to_event_mapper_ids is None:
                self.xcom_to_event_mapper_ids = []
            else:
                for key in self.xcom_to_event_mapper_ids:
                    value = task_instance.xcom_pull(key=key)
                    if value is not None:
                        xcom_context_data[key] = value
            try:
                self.operator.prepare_for_execution().execute(context)
            except Exception as e1:
                if type(e1) == AttributeError:
                    try:
                        self.operator.execute(context)
                    except Exception as e2:
                        LOGGER.error(e2)
                        raise e2
                else:
                    LOGGER.error(e1)
                    raise e1
        except Exception as e:
            LOGGER.error("Exception : %s", e)
            handle_span_failure(e, self.span_context, self.span_uid, try_number, xcom_context_data, max_retries)
        else:
            handle_span_success(self.span_context, self.span_uid, try_number, xcom_context_data)

    def set_downstream(self, task_or_task_list) -> None:
        super().set_downstream(task_or_task_list)

    def set_upstream(self, task_or_task_list) -> None:
        super().set_upstream(task_or_task_list)



from datetime import datetime, timedelta
import datetime
from acceldata_airflow_sdk.dag import DAG
from airflow.operators.python_operator import PythonOperator
import os
from acceldata_airflow_sdk.operators.torch_initialiser_operator import TorchInitializer
from acceldata_sdk.models.job import CreateJob, JobMetadata, Node
from acceldata_sdk.models.pipeline import PipelineMetadata
from acceldata_airflow_sdk.decorators.job import job

job_metadata = JobMetadata('sangeeta@acceldata.io', 'DR', ' Data Pipeline')
pipeline_uid = "job_deco_success_dag_2_retries"
pipeline_name = "job_deco_success_dag_2_retries"

os.environ["no_proxy"] = "*"


@job(job_uid='get_config',
     inputs=[Node(asset_uid="Snowflake_stg.SNOWFLAKE_SAMPLE_DATA.TPCDS_SF100TCL.CALL_CENTER")],
     outputs=[Node(job_uid="process_data"),
              Node(asset_uid="Snowflake_stg.SNOWFLAKE_SAMPLE_DATA.TPCDS_SF100TCL.CUSTOMER")],
     metadata=job_metadata,
     span_uid='get_config.span')
def get_config(**context):
    config = {
        'param1': 'value1',
        'param2': 'value2'
    }
    return config


@job(job_uid='process_data',
     inputs=[Node(job_uid="get_config")],
     outputs=[Node(asset_uid="Snowflake_stg.SNOWFLAKE_SAMPLE_DATA.TPCDS_SF100TCL.REASON")],
     metadata=job_metadata,
     span_uid='process_data.span')
def process_data(**context):
    """
    Process data based on configuration.
    """
    """
    Your task logic goes here.
    """
    # Example: Simulating a task failure
    try_number = context['ti'].try_number
    if try_number <= 2:
        raise ValueError("Simulated error occurred!")
    print("I am successful in the last try")


default_args = {
    'owner': 'airflow',
    'depends_on_past': False,
    'email_on_failure': False,
    'email_on_retry': False,
    'retries': 2,  # Number of retries
    'retry_delay': timedelta(minutes=2)  # Retry delay
}

dag = DAG(
    dag_id="job_deco_success_dag_2_retries", catchup=False,
    default_args=default_args,
    schedule=None, description="Reviews Data Pipeline",
    start_date=datetime.datetime(2024, 4, 24),
    tags=["ad-pipeline-example"]
)

# Define a task to fetch configuration
# Define your task with PythonOperator
# Initialize ADOC
torch_initializer_task = TorchInitializer(
    task_id='torch_pipeline_initializer',
    pipeline_uid=pipeline_uid,
    pipeline_name=pipeline_name,
    dag=dag,
    # Torch connection id goes here
    connection_id='acceldata-poc',
    # optional root span: span_name=pipeline_uid + '_' + 'root_span',
    meta=PipelineMetadata(owner='Demo', team='demo_team', codeLocation='...')
)

config_task = PythonOperator(
    task_id='config_task',
    python_callable=get_config,
    provide_context=True,
    dag=dag
)
process_data_task = PythonOperator(
    task_id='process_data_task',
    provide_context=True,
    python_callable=process_data,
    dag=dag
)

torch_initializer_task >> config_task >> process_data_task



from airflow.models.baseoperator import BaseOperator
from acceldata_sdk.torch_client import TorchClient
from acceldata_airflow_sdk.initialiser import torch_credentials
from acceldata_sdk.constants import FailureStrategy
from acceldata_airflow_sdk.utils.airflow_job_utils import CONNECTION_ID, get_dag_run_pipeline_run_id
from acceldata_sdk.models.common_types import PolicyExecutionRequest

import logging
LOGGER = logging.getLogger("tasks")


class ExecutePolicyOperator(BaseOperator):
    """
    Description:
        ExecutePolicyOperator is used to execute a policy by passing policy_type and policy_id.
        It will return only after the execution ends if sync is set to True.

        :param policy_type: (String) Type of policy to be executed
        :param policy_id: (String) id of the policy to be executed
        :param incremental: (bool) optional Set it to True if full execution has to be done
        :param sync: (bool) optional Set it to True to execute policy in synchronous mode
        :param failure_strategy: (enum) optional Set it to decide if it should fail at error,
            fail at warning or never fail
    """

    def __init__(self, *, policy_type, policy_id, sync, incremental=False,
                 failure_strategy: FailureStrategy = FailureStrategy.DoNotFail,
                 policy_execution_request: PolicyExecutionRequest = None, **kwargs):
        """
        :param policy_type: (PolicyType) Type of policy to be executed
        :param policy_id: (String) id of the policy to be executed
        :param incremental: (bool) optional Set it to True if full execution has to be done
        :param sync: (bool) optional Set it to True to execute policy in synchronous mode
        :param failure_strategy: (enum) optional Set it to decide if it should fail at error,
            fail at warning or never fail
        :param policy_execution_request: (PolicyExecutionRequest) An optional parameter that allows you to provide
        additional options for executing the policy. It is an instance of the PolicyExecutionRequest modules class,
        which contains various properties that can be used to customize the policy execution, such as `executionType`,
        `markerConfigs`, `ruleItemSelections`, and more.

        Example:
        from acceldata_sdk.constants import RuleExecutionStatus, FailureStrategy, PolicyType
        from acceldata_sdk.common import Executor
        operator_task = ExecutePolicyOperator(
            task_id='torch_pipeline_operator_test',
            policy_type=PolicyType.DATA_QUALITY,
            policy_id=46,
            sync=True,
            failure_strategy=FailureStrategy.FailOnError,
            policy_execution_request=policy_execution_request
            dag=dag
        )

        In case you need to query the status in another task you need to pull the execution id from xcom by passing
        the policy name in the {policy_name}_execution_id. In this example the policy name of policy_id 46 is 'policy_with_email'

        After getting the execution_id you need to create object of Executor by passing policy_type and
        torch_client object and call get_result using the execution_id.

        def operator_result(**context):
            xcom_key = {policy_type.name}_{policy_id}_execution_id'
            task_instance = context['ti']
            # get the policy_name and execution id - then pull from xcom
            execution_id = task_instance.xcom_pull(key=xcom_key)
            if execution_id is not None:
                torch_client = TorchClient(**torch_credentials)
                result = torch_client.get_policy_execution_result(policy_type=PolicyType.DATA_QUALITY, execution_id=execution_id)
        """
        super().__init__(**kwargs)
        self.policy_type = policy_type
        self.policy_id = policy_id
        self.incremental = incremental
        self.failure_strategy = failure_strategy
        self.sync = sync
        self.policy_execution_request = policy_execution_request

    def execute(self, context):
        task_instance = context['ti']
        conn_id = task_instance.xcom_pull(key=CONNECTION_ID)
        torch_client = TorchClient(**torch_credentials(conn_id))
        pipeline_run_id = get_dag_run_pipeline_run_id(task_instance)
        execution_return = torch_client.execute_policy(
            policy_type=self.policy_type,
            policy_id=self.policy_id,
            sync=self.sync,
            incremental=self.incremental,
            failure_strategy=self.failure_strategy,
            pipeline_run_id=pipeline_run_id,
            policy_execution_request=self.policy_execution_request
        )
        if execution_return.id is not None:
            xcom_key = f'{self.policy_type.name}_{self.policy_id}_execution_id'
            task_instance = context['ti']
            # get the policy_name and execution id - then push them in xcom
            task_instance.xcom_push(key=xcom_key, value=execution_return.id)




import functools
import logging
from acceldata_airflow_sdk.utils.torch_client import TorchDAGClient
from acceldata_airflow_sdk.utils.airflow_job_utils import PIPELINE_UID_XCOM, get_dag_run_pipeline_run_id, CONNECTION_ID, \
    get_task_instance, pull_xcom_data, is_span_not_found, send_span_start_generic_event, create_new_span, \
    generate_span_uid, handle_span_failure, handle_span_success, fetch_xcom_data, log_tries

from acceldata_sdk.errors import APIError

LOGGER = logging.getLogger("airflow.task")


def span(span_uid=None, associated_job_uids=None, xcom_to_event_mapper_ids=None):
    """
    Description:
        Used to decorate function for which you need span in side your pipeline. Just decorate your function with `span`
    :param xcom_to_event_mapper_ids: xcom pull ids that you want to send with span event
    :param associated_job_uids: list of string
    :param span_uid: uid of the span

    Example:

    @span(span_uid='customer.orders.datagen.span')
    def function(**context)

    """

    def decorator_span(func):
        @functools.wraps(func)
        def wrapper_span(*args, **kwargs):
            span_context, xcom_context_data, try_number, temp_span_uid, max_retries, pipeline_run = None, None, None, None, None, None
            try:
                LOGGER.info("Sending Span Start Event")
                task_instance = get_task_instance(kwargs, func)
                pipeline_uid_ = pull_xcom_data(task_instance, PIPELINE_UID_XCOM)
                conn_id = pull_xcom_data(task_instance, CONNECTION_ID)
                xcoms = xcom_to_event_mapper_ids
                try_number = task_instance.try_number
                max_retries = task_instance.task.retries
                log_tries(try_number, max_retries)
                parent_span_context = pull_xcom_data(task_instance, 'parent_span_context')
                if parent_span_context is None:
                    LOGGER.debug('Sending new request to ADOC Pipeline service to get parent span context')
                    client = TorchDAGClient(conn_id)
                    pipeline_run_id = get_dag_run_pipeline_run_id(task_instance)
                    parent_span_context = client.get_root_span(pipeline_uid=pipeline_uid_,
                                                               pipeline_run_id=pipeline_run_id)
                    pipeline = client.get_pipeline(pipeline_uid_)
                    pipeline_run = pipeline.get_run(pipeline_run_id)
                    # task_instance.xcom_push(key="parent_span_context", value=parent_span_context)
                else:
                    LOGGER.debug('using xcom to get parent span context to send span event')
                associated_job_uids_with_span = associated_job_uids
                if span_uid is None:
                    temp_span_uid = generate_span_uid(func, task_instance, span_uid)
                else:
                    temp_span_uid = span_uid
                try:
                    span_context = pipeline_run.get_span(temp_span_uid)
                    LOGGER.info("Span already exists with child_span_uid : %s . Not creating the span again.",
                                temp_span_uid)
                    # No need to create a span, emit generic events related to the retry here
                    send_span_start_generic_event(span_context, try_number, span_uid, xcoms)
                except APIError as e:
                    LOGGER.warning("APIError : %s", e)
                    if is_span_not_found(e):
                        LOGGER.info("Span does not exist with child_span_uid : %s. Creating new child span. ",
                                    temp_span_uid)
                        span_context = create_new_span(parent_span_context, temp_span_uid, xcoms, try_number,
                                                       associated_job_uids_with_span)
                xcom_context_data = {}
                if xcoms is None:
                    xcoms = []
                else:
                    for key in xcoms:
                        value = task_instance.xcom_pull(key=key)
                        if value is not None:
                            xcom_context_data[key] = value
                kwargs['span_context_parent'] = span_context
                LOGGER.info('Xcom context data %s', xcom_context_data)
                func(*args, **kwargs)
            except Exception as e:
                handle_span_failure(e, span_context, span_uid, try_number, xcom_context_data, max_retries)
            else:
                handle_span_success(span_context, span_uid, try_number, xcom_context_data)

        return wrapper_span

    return decorator_span



import functools
import logging
from acceldata_airflow_sdk.utils.airflow_job_utils import create_job_span, log_job_details, get_task_instance, \
    generate_job_uid, handle_span_success, create_context_job, is_span_not_found, create_new_span, \
    handle_span_failure, log_tries

LOGGER = logging.getLogger("airflow.task")


def job(job_uid=None, span_uid=None, inputs=[], outputs=[], metadata=None, xcom_to_event_mapper_ids=None,
        bounded_by_span=True):
    """
    Description:
    Use this decorator to create functional node (job) in your pipeline and crate span for your function inside your
     pipeline.
    :param job_uid: optional job uid of the pipeline. If not provided default job_id will bre created using dagname,
            task_id  and function name of function being wrapped
    :param span_uid: optional uid of the span. If not passed job_uid will get used as span_uid
    :param inputs: input arrays of the task
    :param outputs: output array of the job
    :param metadata: metadata of the job
    :param xcom_to_event_mapper_ids: xcom pull ids that you want to send with span event
    :param bounded_by_span: optional True by default. Set True if you want span to be created for the job as well


    Example:
    @job(job_uid='customer.order.join.job',
        inputs=[Node(asset_uid='POSTGRES_LOCAL_DS.pipeline.pipeline.orders'), Node(asset_uid='POSTGRES_LOCAL_DS.pipeline.pipeline.customers')] ,
        outputs=[Node(asset_uid='POSTGRES_LOCAL_DS.pipeline.pipeline.customer_orders')],
        metadata=JobMetadata('name', 'team', 'code_location'),
        span_uid='customer.orders.datagen.span',
        bounded_by_span=True)
    def function(**context)

    """

    def decorator_job(func):
        @functools.wraps(func)
        def wrapper_job(*args, **kwargs):
            log_job_details(inputs, outputs)
            task_instance = get_task_instance(kwargs, func)
            temp_job_uid = generate_job_uid(func, task_instance, job_uid)
            context_job = create_context_job("decorator", temp_job_uid, "operator", str(func))
            span_uid_temp, kwargs, xcom_context_data = create_job_span(
                task_instance=task_instance,
                job_uid=temp_job_uid,
                inputs=inputs,
                outputs=outputs,
                metadata=metadata,
                context_job=context_job,
                bounded_by_span=bounded_by_span,
                xcom_to_event_mapper_ids=xcom_to_event_mapper_ids,
                span_uid=span_uid,
                kwargs=kwargs
            )
            span_context = kwargs.get('span_context_parent', None)
            try_number = task_instance.try_number
            max_retries = task_instance.task.retries
            log_tries(try_number, max_retries)
            try:
                func(*args, **kwargs)
            except Exception as e:
                handle_span_failure(e, span_context, span_uid_temp, try_number, xcom_context_data, max_retries)
            else:
                handle_span_success(span_context, span_uid_temp, try_number, xcom_context_data)

        return wrapper_job

    return decorator_job



from datetime import datetime, timedelta
# from airflow import DAG
import datetime
from acceldata_sdk.models.job import CreateJob, JobMetadata, Node
from acceldata_sdk.models.pipeline import PipelineMetadata
from acceldata_sdk.events.generic_event import GenericEvent
from acceldata_airflow_sdk.dag import DAG
from airflow.operators.python import PythonOperator
from acceldata_airflow_sdk.decorators.job import job
from acceldata_airflow_sdk.operators.torch_initialiser_operator import TorchInitializer
import os

os.environ["no_proxy"] = "*"

job_metadata = JobMetadata('sangeeta@acceldata.io', 'DR', ' Data Pipeline')
pipeline_uid = "job_deco_success_dag_0_retries"
pipeline_name = "job_deco_success_dag_0_retries"

timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S")


@job(job_uid='get_config',
     inputs=[Node(asset_uid="Snowflake_stg.SNOWFLAKE_SAMPLE_DATA.TPCDS_SF100TCL.CALL_CENTER")],
     outputs=[Node(job_uid="process_data"),
              Node(asset_uid="Snowflake_stg.SNOWFLAKE_SAMPLE_DATA.TPCDS_SF100TCL.CUSTOMER")],
     metadata=job_metadata,
     span_uid='get_config.span')
def get_config(**context):
    config = {
        'param1': 'value1',
        'param2': 'value2'
    }
    return config


@job(job_uid='process_data',
     inputs=[Node(job_uid="get_config")],
     outputs=[Node(asset_uid="Snowflake_stg.SNOWFLAKE_SAMPLE_DATA.TPCDS_SF100TCL.REASON")],
     metadata=job_metadata,
     span_uid='process_data.span')
def process_data(**context):
    """
    Process data based on configuration.
    """
    """
    Your task logic goes here.
    """
    # Example: Processed data
    processed_data = {
        'param1': 'processed_value1',
        'param2': 'processed_value2'
    }
    return processed_data


default_args = {
    "owner": "Sangeeta Airflow", 'retries': 0, "retry_delay": timedelta(minutes=3), "depends_on_past": False,
    "email": ["sangeeta@accceldata.io"], "email_on_failure": False, "email_on_retry": False
}

dag = DAG(
    dag_id="job_deco_success_dag_0_retries", catchup=False, default_args=default_args,
    schedule=None, description="Reviews Data Pipeline",
    max_active_runs=2, start_date=datetime.datetime(2024, 4, 24), tags=["ad-pipeline-example"]
)

# Initialize ADOC
torch_initializer_task = TorchInitializer(
    task_id='torch_pipeline_initializer',
    pipeline_uid=pipeline_uid,
    pipeline_name=pipeline_name,
    dag=dag,
    # Torch connection id goes here
    connection_id='acceldata-poc',
    # optional root span: span_name=pipeline_uid + '_' + 'root_span',
    meta=PipelineMetadata(owner='Demo', team='demo_team', codeLocation='...')
)

config_task = PythonOperator(
    task_id='config_task',
    provide_context=True,
    python_callable=get_config,
    dag=dag)

process_data_task = PythonOperator(
    task_id='process_data_task',
    provide_context=True,
    python_callable=process_data,
    dag=dag
)

torch_initializer_task >> config_task >> process_data_task



from datetime import datetime, timedelta
# from airflow import DAG
import datetime
from airflow.operators.bash import BashOperator
from acceldata_sdk.models.job import CreateJob, JobMetadata, Node
from acceldata_sdk.models.pipeline import PipelineMetadata
from acceldata_sdk.events.generic_event import GenericEvent
from acceldata_airflow_sdk.dag import DAG
from acceldata_airflow_sdk.operators.torch_initialiser_operator import TorchInitializer
from acceldata_airflow_sdk.operators.span_operator import SpanOperator
from acceldata_airflow_sdk.operators.job_operator import JobOperator
from acceldata_airflow_sdk.operators.execute_policy_operator import ExecutePolicyOperator
from acceldata_sdk.constants import FailureStrategy, PolicyType

import os

os.environ["no_proxy"] = "*"

job_metadata = JobMetadata('sangeeta@acceldata.io', 'DR', ' Data Pipeline')
pipeline_uid = "execute_policy_operator_dag"
pipeline_name = "execute_policy_operator_dag"

timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S")

default_args = {
    "owner": "Sangeeta Airflow", 'retries': 0, "depends_on_past": False,
    "email": ["sangeeta@accceldata.io"], "email_on_failure": False, "email_on_retry": False
}

dag = DAG(
    dag_id="execute_policy_operator_dag", catchup=False, default_args=default_args,
    schedule=None, description="execute_policy_operator_dag", start_date=datetime.datetime(2024, 4, 24),
    tags=["ad-pipeline-example"]
)

# Initialize ADOC
torch_initializer_task = TorchInitializer(
    task_id='torch_pipeline_initializer',
    pipeline_uid=pipeline_uid,
    pipeline_name=pipeline_name,
    dag=dag,
    # Torch connection id goes here
    connection_id='acceldata-poc',
    # optional root span: span_name=pipeline_uid + '_' + 'root_span',
    meta=PipelineMetadata(owner='Demo', team='demo_team', codeLocation='...')
)

execute_policy_task = ExecutePolicyOperator(
    task_id='execute_policy_operator',
    policy_type=PolicyType.RECONCILIATION,
    policy_id=269722,
    sync=False,
    failure_strategy=FailureStrategy.DoNotFail
)

execute_policy_operator = JobOperator(
    task_id='execute_policy_operator_task',
    job_uid='execute_policy_operator',
    inputs=[Node(asset_uid="Snowflake_stg.SNOWFLAKE_SAMPLE_DATA.TPCDS_SF100TCL.CALL_CENTER")],
    outputs=[Node(job_uid="bash_task_2"),
             Node(asset_uid="Snowflake_stg.SNOWFLAKE_SAMPLE_DATA.TPCDS_SF100TCL.CUSTOMER")],
    metadata=JobMetadata('name', 'team', 'code_location'),
    operator=execute_policy_task,
    dag=dag,
    xcom_to_event_mapper_ids=['run_id', 'event_id']
)

torch_initializer_task >> execute_policy_operator



import functools

import logging

from acceldata_airflow_sdk.utils.callback import on_dag_success_callback, on_dag_failure_callback

LOGGER = logging.getLogger("airflow.task")


def handle_dag_callback(func):
    """
    Used to decorate callback function in side torch dag implementation.
    :param func: function to decorate
    """
    @functools.wraps(func)
    def wrapper_callback(*args, **kwargs):
        try:
            LOGGER.info("Decorator callback - start")
            context = None
            if len(args) > 0:
                context = args[0]
            if context is not None:
                if context['reason'] == 'success':
                    on_dag_success_callback(context=context)
                else:
                    on_dag_failure_callback(context=context)
            func(*args, **kwargs)
        except Exception as e:
            LOGGER.error("Decorator callback failed")
            exception = e.__dict__
            LOGGER.error(exception)
            raise e
        else:
            LOGGER.info("Decorator callback - end")

    return wrapper_callback



from datetime import datetime
from airflow.models.baseoperator import BaseOperator
from acceldata_airflow_sdk.utils.airflow_job_utils import create_job_span, log_job_details, get_task_instance, \
    generate_job_uid, handle_span_success, create_context_job, is_span_not_found, create_new_span, \
    handle_span_failure, create_context_job, generate_job_operator_uid, log_tries
import logging

LOGGER = logging.getLogger("airflow.task")


class JobOperator(BaseOperator):
    """
    Description:
        Used to create Job and send span start and end event for any std airflow operator. Just wrap your operator with
        Job operator. Make sure you do not add your task in dag(dag parameter should not be specified in the
        operator being wrapped by job operator). If you wrap it using Job operator, will take care of that task
         operator.

    You need to add extra parameter mentioned below. Other parameters will be same as std airflow base operator's
     parameters

    :param operator: std task operator defined
    :param job_uid: optional job uid of the pipeline. If not provided default job_id will bre created using dagname,
        task_id of the base operator and operator name of base operator being wrapped
    :param inputs: input arrays of the task
    :param outputs: output array of the job
    :param metadata: metadata of the job
    :param span_uid: span uid for the task. If not passed job_uid will get used as span_uid
    :param xcom_to_event_mapper_ids: list of xcom keys. Used to send xcom variables in span event job
    :param bounded_by_span: optional True if you want to create a span for the current task
    """

    def __init__(self, *, operator: BaseOperator, inputs=[], outputs=[], job_uid: str = None, metadata=None,
                 span_uid: str = None,
                 xcom_to_event_mapper_ids=None, bounded_by_span=True, **kwargs):
        """
        You need to add extra parameter mentioned below. Other parameters will be same as std airflow base operator's parameters
        :param operator: std task operator defined
        :param job_uid: optional job uid of the pipeline. If not provided default job_id will bre created using dagname,
            task_id of the base operator and operator name of base operator being wrapped
        :param inputs: input arrays of the task
        :param outputs: output array of the job
        :param metadata: optional metadata of the job
        :param span_uid: optional span uid for the task. If not passed job_uid will get used as span_uid
        :param xcom_to_event_mapper_ids: list of xcom keys. Used to send xcom variables in span event
        :param bounded_by_span: optional True if you want to create a span for the current task
        Example :

        --> Defined std operator.

        postgres_operator = PostgresOperator(
            task_id="task_name",
            postgres_conn_id='example_db',
            sql="select * from information_schema.attributes",
        )


        --> To create Job and wrap operator with span. Assign this to your dag (not your std operator)

        job_operator = JobOperator(
            task_id='task_name',
            job_uid='customer.order.join.job',
            inputs=[Node('POSTGRES_LOCAL_DS.pipeline.pipeline.orders'), Node('POSTGRES_LOCAL_DS.pipeline.pipeline.customers')] ,
            outputs=[Node('POSTGRES_LOCAL_DS.pipeline.pipeline.customer_orders')],
            metadata=JobMetadata('name', 'team', 'code_location'),
            span_uid='span.uid',
            operator=postgres_operator,
            dag=dag,
            bounded_by_span=True
        )

        """
        if kwargs.get("provide_context"):
            kwargs.pop('provide_context', None)
        super().__init__(**kwargs)
        self.operator = operator
        self.pipeline_uid = None
        self.job_uid = job_uid
        self.span_uid = span_uid
        self.inputs = inputs
        self.outputs = outputs
        self.metadata = metadata
        self.xcom_to_event_mapper_ids = xcom_to_event_mapper_ids
        self.parent_span_ctxt = None
        self.span_context = None
        self.bounded_by_span = bounded_by_span

    def execute(self, context):
        xcom_context_data = None
        context_job = create_context_job("operator", self.job_uid, "operator", str(self.operator))
        task_instance = context['ti']
        if self.job_uid is None:
            self.job_uid = generate_job_operator_uid(type(self.operator).__name__, self.operator.task_id)
        self.span_uid, context, xcom_context_data = create_job_span(
            task_instance=task_instance,
            job_uid=self.job_uid,
            inputs=self.inputs,
            outputs=self.outputs,
            metadata=self.metadata,
            context_job=context_job,
            bounded_by_span=self.bounded_by_span,
            xcom_to_event_mapper_ids=self.xcom_to_event_mapper_ids,
            span_uid=self.span_uid,
            kwargs=context
        )
        self.span_context = context.get('span_context_parent', None)
        try_number = task_instance.try_number
        max_retries = task_instance.task.retries
        log_tries(try_number, max_retries)
        try:
            try:
                self.operator.prepare_for_execution().execute(context)
            except Exception as e1:
                if type(e1) == AttributeError:
                    try:
                        self.operator.execute(context)
                    except Exception as e2:
                        LOGGER.error(e2)
                        raise e2
                else:
                    LOGGER.error(e1)
                    raise e1
        except Exception as e:
            LOGGER.error("Send span end failure event")
            exception = e.__dict__
            LOGGER.error(exception)
            if self.span_context is not None:
                handle_span_failure(e, self.span_context, self.span_uid, try_number, xcom_context_data, max_retries)
        else:
            handle_span_success(self.span_context, self.span_uid, try_number, xcom_context_data)

    def set_downstream(self, task_or_task_list) -> None:
        super().set_downstream(task_or_task_list)

    def set_upstream(self, task_or_task_list) -> None:
        super().set_upstream(task_or_task_list)



from setuptools import setup, find_packages

classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Operating System :: OS Independent',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python :: 3'
]

requirements = [
    "acceldata-sdk>=3.12.1",
    'requests',
    'dataclasses;python_version<"3.7"',
    'typing;python_version<"3.5"'
]

setup(
    name='acceldata_airflow_sdk',
    version='3.15.0',
    description=open('README.txt').read(),
    long_description=open('README.md').read() + '\n\n' + open('CHANGELOG.txt').read(),
    long_description_content_type="text/markdown",
    url='',
    author='acceldata',
    author_email='apisupport@acceldata.io',
    license='MIT License',
    classifiers=classifiers,
    keywords='acceldata-airflow',
    packages=find_packages(),
    install_requires=requirements
)



from datetime import datetime, timedelta
# from airflow import DAG
import datetime
from airflow.operators.bash import BashOperator
from acceldata_sdk.models.job import CreateJob, JobMetadata, Node
from acceldata_sdk.models.pipeline import PipelineMetadata
from acceldata_airflow_sdk.dag import DAG
from acceldata_airflow_sdk.operators.torch_initialiser_operator import TorchInitializer
import os
from acceldata_airflow_sdk.operators.span_operator import SpanOperator
from acceldata_airflow_sdk.operators.job_operator import JobOperator

os.environ["no_proxy"] = "*"

job_metadata = JobMetadata('sangeeta@acceldata.io', 'DR', ' Data Pipeline')
pipeline_uid = "span_operator_success_dag_2_retries"
pipeline_name = "span_operator_success_dag_2_retries"

timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S")


def get_bash_command():
    return """
    ATTEMPT_FILE="/tmp/airflow_bash_operator_attempt_count.txt"
    if [ ! -f "$ATTEMPT_FILE" ]; then
      echo 0 > "$ATTEMPT_FILE"
    fi
    ATTEMPT_COUNT=$(cat "$ATTEMPT_FILE")
    ATTEMPT_COUNT=$((ATTEMPT_COUNT + 1))
    echo $ATTEMPT_COUNT > "$ATTEMPT_FILE"
    if [ "$ATTEMPT_COUNT" -eq 2 ]; then
      echo "Success on attempt $ATTEMPT_COUNT"
      rm "$ATTEMPT_FILE"  # Cleanup the attempt file after success
      exit 0
    else
      echo "Failing attempt $ATTEMPT_COUNT"
      exit 1
    fi
    """


default_args = {
    "owner": "Sangeeta Airflow", 'retries': 2, "retry_delay": timedelta(minutes=2), "depends_on_past": False,
    "email": ["sangeeta@accceldata.io"], "email_on_failure": False, "email_on_retry": False
}

dag = DAG(
    dag_id="span_operator_success_dag_2_retries", catchup=False, default_args=default_args,
    schedule=None, description="span_operator_success_dag_2_retries",
    max_active_runs=2, start_date=datetime.datetime(2024, 4, 24), tags=["ad-pipeline-example"]
)

# Initialize ADOC
torch_initializer_task = TorchInitializer(
    task_id='torch_pipeline_initializer',
    pipeline_uid=pipeline_uid,
    pipeline_name=pipeline_name,
    dag=dag,
    # Torch connection id goes here
    connection_id='acceldata-poc',
    # optional root span: span_name=pipeline_uid + '_' + 'root_span',
    meta=PipelineMetadata(owner='Demo', team='demo_team', codeLocation='...')
)

bash_task_1 = BashOperator(
    task_id="bash_task_1",
    bash_command='echo "Here is the message: \'{{ dag_run.conf["message"] if dag_run.conf else "" }}\'"',
)

bash_operator_1 = JobOperator(
    task_id='bash_task_1',
    job_uid='bash_task_1',
    inputs=[Node(asset_uid="Snowflake_stg.SNOWFLAKE_SAMPLE_DATA.TPCDS_SF100TCL.CALL_CENTER")],
    outputs=[Node(job_uid="bash_operator_retry"),
             Node(asset_uid="Snowflake_stg.SNOWFLAKE_SAMPLE_DATA.TPCDS_SF100TCL.CUSTOMER")],
    metadata=JobMetadata('name', 'team', 'code_location'),
    operator=bash_task_1,
    dag=dag,
    xcom_to_event_mapper_ids=['run_id', 'event_id']
)

# Define the BashOperator using the generated command
bash_operator_retry = BashOperator(
    task_id='bash_operator_retry',
    bash_command=get_bash_command()
)

bash_operator_2 = SpanOperator(
    task_id='bash_operator_retry',
    span_uid='bash_operator_retry.span',
    operator=bash_operator_retry,
    dag=dag,
    xcom_to_event_mapper_ids=['run_id', 'event_id'],
    associated_job_uids=['bash_task_2']
)

torch_initializer_task >> bash_operator_1 >> bash_operator_2



from acceldata_airflow_sdk.utils.torch_client import TorchDAGClient
from acceldata_sdk.models.job import CreateJob, JobMetadata
from acceldata_sdk.events.generic_event import GenericEvent
from acceldata_sdk.errors import APIError
import json
from datetime import datetime
import logging

LOGGER = logging.getLogger(__name__)
PIPELINE_UID_XCOM = 'pipeline_uid_ff069534-5069-45b1-b737-aea6229db4bf'
CONNECTION_ID = 'CONNECTION_ID'


def get_dag_run_pipeline_run_id(task_instance):
    return task_instance.xcom_pull(key=f'{task_instance.dag_id}_pipeline_run_id')


def create_job_span(task_instance, job_uid, inputs, outputs, metadata, context_job, bounded_by_span,
                    xcom_to_event_mapper_ids, span_uid, kwargs):
    local_span_context = None
    span_uid_temp = None
    xcom_context_data = None
    try:
        xcom_context_data = {}
        span_uid_temp = span_uid
        conn_id = pull_xcom_data(task_instance, CONNECTION_ID)
        pipeline_uid_ = pull_xcom_data(task_instance, PIPELINE_UID_XCOM)
        client = TorchDAGClient(conn_id)
        pipeline = client.get_pipeline(pipeline_uid_)
        pipeline_run_id = get_dag_run_pipeline_run_id(task_instance)
        pipeline_run = pipeline.get_run(pipeline_run_id)
        LOGGER.info("Creating job")
        job = create_job(pipeline, job_uid, inputs, outputs, metadata, context_job, pipeline_run)
    except Exception as e:
        handle_job_creation_failure(e)
    else:
        LOGGER.info("Successfully created job.")
        if bounded_by_span:
            LOGGER.info('Sending Span Start event')
            try:
                xcoms = xcom_to_event_mapper_ids
                parent_span_context = pull_xcom_data(task_instance, "parent_span_context")
                if parent_span_context is None:
                    LOGGER.debug('Sending new request to ADOC Pipeline service to get parent span context')
                    parent_span_context = client.get_root_span(pipeline_uid=pipeline_uid_)
                else:
                    LOGGER.debug('Using xcom to get parent span context to send span event')
                associated_job_uids = [job_uid]
                if span_uid is None:
                    span_uid_temp = job_uid
                try_number = task_instance.try_number
                # 3. Handle the case when the span already exists (retry)
                try:
                    local_span_context = pipeline_run.get_span(span_uid_temp)
                    LOGGER.info("Span already exists with child_span_uid : %s . Not creating the span again.",
                                span_uid_temp)
                    # No need to create a span, emit generic events related to the retry here
                    send_span_start_generic_event(local_span_context, try_number, span_uid, xcoms)
                except APIError as e:
                    LOGGER.warn("APIError : %s", e)
                    if is_span_not_found(e):
                        LOGGER.info("Span does not exist with child_span_uid : %s. Creating new child span. ",
                                    span_uid_temp)
                        local_span_context = create_new_span(parent_span_context, span_uid_temp, xcoms, try_number,
                                                             associated_job_uids)
                if xcoms is None:
                    xcoms = []
                else:
                    xcom_context_data = fetch_xcom_data(xcoms, task_instance)
                    LOGGER.info('Xcom context data :: %s ', xcom_context_data)
                kwargs['span_context_parent'] = local_span_context
            except Exception as ex:
                LOGGER.error(f'Span creation failed with exception: {str(ex)}')
    return span_uid_temp, kwargs, xcom_context_data


def is_pipeline_run_ended(pipeline_run):
    if pipeline_run.status == "COMPLETED" or pipeline_run.status == "FAILED" or pipeline_run.status == "ABORTED":
        return True


def log_job_details(inputs, outputs):
    LOGGER.info(f"Job inputs: {inputs}")
    LOGGER.info(f"Job outputs: {outputs}")


def log_tries(task_try_number, max_retries):
    LOGGER.info("try_number inside task failure/success callback %s", task_try_number)
    LOGGER.info("max_retries %s", max_retries)


def handle_job_creation_failure(e):
    LOGGER.error("Error in creating job")
    exception = e.__dict__
    LOGGER.error(exception)
    raise e


def get_task_instance(kwargs, func):
    if kwargs is None or kwargs.get('ti', None) is None:
        raise Exception(f'Please pass context to function:  {str(func.__name__)}')
    return kwargs['ti']


def generate_job_uid(func, task_instance, job_uid):
    if job_uid is None:
        fname_six = '{:.6}'.format(func.__name__)
        job_uid = f'{task_instance.task_id}_{fname_six}'
    return job_uid


def generate_job_operator_uid(operator_name, task_id):
    opame_six = '{:.6}'.format(operator_name)
    job_uid = f'{task_id}_{opame_six}'
    return job_uid


def generate_span_uid(func, task_instance, span_uid):
    if span_uid is None:
        fname_six = '{:.6}'.format(func.__name__)
        span_uid = f'{task_instance.task_id}_{fname_six}.span'
    return span_uid


def generate_span_operator_uid(operator_name, task_id):
    opame_six = '{:.6}'.format(operator_name)
    span_uid = f'{task_id}_{opame_six}.span'
    return span_uid


def create_context_job(type, job_uid, function_key, func_value):
    return {
        'job': 'adoc_job_' + type,
        'time': str(datetime.now()),
        'uid': job_uid,
        function_key: str(func_value)
    }


def handle_span_failure(e, temp_span_context, span_uid, try_number, xcom_context_data, retries):
    LOGGER.error("Handling task failure event")
    exception = e.__dict__
    LOGGER.error(exception)
    if temp_span_context is not None:
        send_span_failure_generic_event(temp_span_context, try_number, span_uid, e)
        # Only end the span if this is the final try
        if try_number > retries:
            LOGGER.info("Sending Span Failed event for the last attempt")
            temp_span_context.failed(context_data=xcom_context_data)
        raise e
    else:
        LOGGER.warning("SpanContext is unavailable. Not able to send Span failure generic event.")


def handle_span_success(temp_span_context, span_uid_temp, try_number, xcom_context_data):
    LOGGER.info("Sending Span End event with status Success")
    if temp_span_context is not None:
        send_span_success_generic_event(temp_span_context, try_number, span_uid_temp)
        temp_span_context.end(context_data=xcom_context_data)
    else:
        LOGGER.warning("SpanContext is unavailable. Not able to send Span Success generic event.")


def send_span_success_generic_event(temp_span_context, try_number, span_uid_temp):
    event_uid = f'{span_uid_temp}.job.end.try_{try_number}.event'
    job_end_event = GenericEvent(context_data={
        'time': str(datetime.now()),
        'try_number': try_number
    }, event_uid=event_uid)
    LOGGER.info("Sending job end event %s for the child span with span_uid %s", job_end_event, span_uid_temp)
    temp_span_context.send_event(job_end_event)


def send_span_start_generic_event(temp_span_context, try_number, span_uid, xcoms=[]):
    LOGGER.info("Sending Span start generic event")
    event_uid = f'{span_uid}.job.start.try_{try_number}.event'
    job_start_event = GenericEvent(context_data={
        'time': str(datetime.now()),
        'try_number': try_number,
        'xcom_to_event_mapper_ids': xcoms},
        event_uid=event_uid)
    LOGGER.info("Sending job retry start event %s for the child span with span_uid %s", job_start_event,
                span_uid)
    temp_span_context.send_event(job_start_event)
    LOGGER.info("Successfully sent job start retry event for the child span with child_span_uid %s",
                span_uid)


def send_span_failure_generic_event(temp_span_context, try_number, span_uid, e):
    LOGGER.debug("Sending Span failure generic event")
    event_uid = f'{span_uid}.job.failed.try_{try_number}.event'
    job_failed_event = GenericEvent(context_data={
        'status': 'error',
        'error_data': str(e),
        'time': str(datetime.now()),
        'try_number': try_number,
        'exception_type': str(type(e).__name__)
    }, event_uid=event_uid)
    temp_span_context.send_event(job_failed_event)


def fetch_xcom_data(xcoms, task_instance):
    xcom_context_data = {}
    for key in xcoms:
        value = pull_xcom_data(task_instance, key)
        if value is not None:
            xcom_context_data[key] = value
    return xcom_context_data


def pull_xcom_data(task_instance, key):
    return task_instance.xcom_pull(key=key)


def initialize_torch_client(task_instance):
    conn_id = pull_xcom_data(task_instance, CONNECTION_ID)
    return TorchDAGClient(conn_id)


def get_pipeline_and_run(client, task_instance):
    pipeline_uid = pull_xcom_data(task_instance, PIPELINE_UID_XCOM)
    pipeline = client.get_pipeline(pipeline_uid)
    pipeline_run_id = get_dag_run_pipeline_run_id(task_instance)
    pipeline_run = pipeline.get_run(pipeline_run_id)
    return pipeline, pipeline_run


def create_job(pipeline, job_uid, inputs, outputs, metadata, context_job, pipeline_run):
    job = CreateJob(
        uid=job_uid,
        name=f'{job_uid} Job',
        pipeline_run_id=pipeline_run.id,
        description=f'{job_uid} created using adoc job decorator',
        inputs=inputs,
        outputs=outputs,
        meta=metadata,
        context=context_job
    )
    return pipeline.create_job(job)


def create_new_span(parent_span_context, span_uid_temp, xcoms, try_number, associated_job_uids):
    LOGGER.info("Creating new span with span_uid %s", span_uid_temp)
    temp_span_context = parent_span_context.create_child_span(
        uid=span_uid_temp,
        context_data={
            'time': str(datetime.now()),
            'xcom_to_event_mapper_ids': xcoms,
            'try_number': try_number
        },
        associatedJobUids=associated_job_uids
    )
    send_span_start_generic_event(temp_span_context, try_number, span_uid_temp, xcoms)
    return temp_span_context


def is_span_not_found(e):
    try:
        error_details = json.loads(e.args[0])
        if error_details.get("errors"):
            for error in error_details["errors"]:
                message = error.get("message", "")
                if "not found" in message.lower() and error.get("status") == 404:
                    return True
        return False
    except:
        return False
